"""
Basic utilities for handling avro schemas
"""
from avro import schema


def parse_schema_from_string(schema_str):
    """Parse a schema given a schema string"""
    return schema.Parse(schema_str)


def parse_schema_from_file(schema_path):
    """Parse a schema from a file path"""
    with open(schema_path) as f:
        return parse_schema_from_string(f.read())



import time
import unittest

from avroschemaserializer.schemaregistry.tests import data_gen
from avroschemaserializer.schemaregistry.tests import mock_registry

from avroschemaserializer.schemaregistry.client import SchemaRegistryClient
from avroschemaserializer.schemaregistry.serializers import Util


class TestCacheSchemaRegistryClient(unittest.TestCase):

    def setUp(self):
        self.server = mock_registry.ServerThread(9001)
        self.server.start()
        time.sleep(1)
        self.client = SchemaRegistryClient('http://127.0.0.1:9001')

    def tearDown(self):
        self.server.shutdown()
        self.server.join()

    def assertLatest(self, meta_tuple, sid, schema, version):
        self.assertNotEqual(sid, -1)
        self.assertNotEqual(version, -1)
        self.assertEqual(meta_tuple[0], sid)
        self.assertEqual(meta_tuple[1], schema)
        self.assertEqual(meta_tuple[2], version)

    def test_register(self):
        parsed = Util.parse_schema_from_string(data_gen.BASIC_SCHEMA)
        client = self.client
        schema_id = client.register('test', parsed)
        self.assertTrue(schema_id == 1)
        self.assertEqual(len(client.id_to_schema), 1)

    def test_multi_subject_register(self):
        parsed = Util.parse_schema_from_string(data_gen.BASIC_SCHEMA)
        client = self.client
        schema_id = client.register('test', parsed)
        self.assertTrue(schema_id == 1)

        # register again under different subject
        dupe_id = client.register('other', parsed)
        self.assertEqual(schema_id, dupe_id)
        self.assertEqual(len(client.id_to_schema), 1)

    def test_dupe_register(self):
        parsed = Util.parse_schema_from_string(data_gen.BASIC_SCHEMA)
        subject = 'test'
        client = self.client
        schema_id = client.register(subject, parsed)
        self.assertTrue(schema_id == 1)
        latest = client.get_latest_schema(subject)

        # register again under same subject
        dupe_id = client.register(subject, parsed)
        self.assertEqual(schema_id, dupe_id)
        dupe_latest = client.get_latest_schema(subject)
        self.assertEqual(latest, dupe_latest)

    def test_getters(self):
        parsed = Util.parse_schema_from_string(data_gen.BASIC_SCHEMA)
        client = self.client
        subject = 'test'
        version = client.get_version(subject, parsed)
        self.assertEqual(version, 1)
        schema = client.get_by_id(1)
        # self.assertEqual(schema, parsed)
        latest = client.get_latest_schema(subject)
        self.assertEqual(latest, (1, schema, 1))

        # register
        schema_id = client.register(subject, parsed)
        latest = client.get_latest_schema(subject)
        version = client.get_version(subject, parsed)
        self.assertLatest(latest, schema_id, parsed, version)

        fetched = client.get_by_id(schema_id)
        self.assertEqual(fetched, parsed)



import os
import random
import os.path

NAMES = ['stefan', 'melanie', 'nick', 'darrel', 'kent', 'simon']
AGES = list(range(1, 10)) + [None]


def get_schema_path(fname):
    dname = os.path.dirname(os.path.realpath(__file__))
    return os.path.join(dname, fname)


def load_schema_file(fname):
    fname = get_schema_path(fname)
    with open(fname) as f:
        return f.read()

BASIC_SCHEMA = load_schema_file('basic_schema.avsc')


def create_basic_item(i):
    return {
        'name' : random.choice(NAMES) + '-' + str(i),
        'number' : random.choice(AGES)
    }

BASIC_ITEMS = map(create_basic_item, range(1, 20))

ADVANCED_SCHEMA = load_schema_file('adv_schema.avsc')


def create_adv_item(i):
    friends = map(create_basic_item, range(1, 3))
    basic = create_basic_item(i)
    basic['friends'] = dict(map(lambda bi: (bi['name'], bi), friends))
    return basic

ADVANCED_ITEMS = map(create_adv_item, range(1, 20))

from avro import schema
from avro.datafile import DataFileReader, DataFileWriter
from avro.io import DatumReader, DatumWriter
import json


def _write_items(base_name, schema_str, items):
    avro_schema = schema.parse(schema_str)
    avro_file = base_name + '.avro'
    with DataFileWriter(open(avro_file, "w"), DatumWriter(), avro_schema) as writer:
        for i in items:
            writer.append(i)
    writer.close
    return (avro_file)


def write_basic_items(base_name):
    return _write_items(base_name, BASIC_SCHEMA, BASIC_ITEMS)


def write_advanced_items(base_name):
    return _write_items(base_name, ADVANCED_SCHEMA, ADVANCED_ITEMS)


def cleanup(files):
    for f in files:
        try:
            os.remove(f)
        except OSError:
            pass

if __name__ == "__main__":
    write_advanced_items("advanced")



try:
  from fastavro.reader import read_data
  from fastavro import dump
  HAS_FAST = True
except:
  HAS_FAST = False

import io
import struct
import avro.io

from avroschemaserializer.schemaregistry.serializers import SerializerError
from avroschemaserializer.schemaregistry.client import ClientError

MAGIC_BYTE = 0


class ContextBytesIO(io.BytesIO):

    """
    Wrapper to allow use of BytesIO via 'with' constructs.
    """

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()
        return False


class MessageSerializer(object):

    """
    A helper class that can serialize and deserialize messages
    that need to be encoded or decoded using the schema registry.

    All encode_* methods return a buffer that can be sent to kafka.
    All decode_* methods expect a buffer received from kafka.
    """

    def __init__(self, registry_client, fast_avro=True):
        self.registry_client = registry_client
        self.id_to_decoder_func = { }
        self.id_to_writers = { }
        self.fast_avro = HAS_FAST

    def _set_subject(self, subject, is_key=False):
        subject_suffix = ('-key' if is_key else '-value')
        # get the latest schema for the subject
        return (subject + subject_suffix)

    def encode_record_with_schema(self, topic, schema, record, is_key=False):
        """
        Given a parsed avro schema, encode a record for the given topic.  The
        record is expected to be a dictionary.

        The schema is registered with the subject of 'topic-value'
        """
        if not isinstance(record, dict):
            raise SerializerError("record must be a dictionary")

        subject = self._set_subject(topic, is_key)

        # register it
        try:
            schema_id = self.registry_client.register(subject, schema)
        except:
            schema_id = None

        if not schema_id:
            message = "Unable to retrieve schema id for subject %s" % (subject)
            raise SerializerError(message)

        if not self.fast_avro:
            self.id_to_writers[schema_id] = avro.io.DatumWriter(schema)

        return self.encode_record_with_schema_id(schema_id, schema, record)

    # subject = topic + suffix
    def encode_record_for_topic(self, topic, record, is_key=False):
        """
        Encode a record for a given topic.

        This is expensive as it fetches the latest schema for a given topic.
        """
        if not isinstance(record, dict):
            raise SerializerError("record must be a dictionary")

        subject = self._set_subject(topic, is_key)

        try:
            schema_id, schema, version = self.registry_client.get_latest_schema(subject)
        except ClientError as e:
            message = "Unable to retrieve schema id for subject %s" % (subject)
            raise SerializerError(message)
        else:
            if not self.fast_avro:
                self.id_to_writers[schema_id] = avro.io.DatumWriter(schema)
            return self.encode_record_with_schema_id(schema_id, schema, record)

    @staticmethod
    def encode_record_with_local_schema(schema, record):
        with ContextBytesIO() as outf:
            outf.write(struct.pack('b', MAGIC_BYTE))
            outf.write(struct.pack('>I', 0))
            encoder = avro.io.BinaryEncoder(outf)
            writer = avro.io.DatumWriter(schema)
            writer.write(record, encoder)
            return outf.getvalue()

    @staticmethod
    def decode_message_with_local_schema(schema, message):
        if len(message) <= 5:
            raise SerializerError("message is too small to decode")

        with ContextBytesIO(message) as payload:
            magic, schema_id = struct.unpack('>bI', payload.read(5))
            if magic != MAGIC_BYTE:
                raise SerializerError("message does not start with magic byte")
            curr_pos = payload.tell()
            avro_reader = avro.io.DatumReader(schema)
            def decoder(p):
                bin_decoder = avro.io.BinaryDecoder(p)
                return avro_reader.read(bin_decoder)
            return decoder(payload)

    def encode_record_with_schema_id(self, schema_id, schema, record):
        """
        Encode a record with a given schema id.  The record must
        be a python dictionary.
        """
        if not isinstance(record, dict):
            raise SerializerError("record must be a dictionary")

        if not self.fast_avro:
            if schema_id not in self.id_to_writers:
                # get the writer + schema
                try:
                    schema = self.registry_client.get_by_id(schema_id)
                    if not schema:
                        raise SerializerError("Schema does not exist")
                    self.id_to_writers[schema_id] = avro.io.DatumWriter(schema)
                except ClientError as e:
                    raise SerializerError("Error fetching schema from registry")

        with ContextBytesIO() as outf:
            # write the header
            # magic byte
            outf.write(struct.pack('b', MAGIC_BYTE))
            # write the schema ID in network byte order (big end)
            outf.write(struct.pack('>I', schema_id))
            if self.fast_avro:
                dump(outf, record, schema.to_json())
            else:
                writer = self.id_to_writers[schema_id]
                encoder = avro.io.BinaryEncoder(outf)
                writer.write(record, encoder)
            return outf.getvalue()

    def get_schema(self, schema_id):
        # fetch from schema reg
        try:
            # first call will cache in the client
            schema = self.registry_client.get_by_id(schema_id)
        except:
            schema = None

        if not schema:
            err = "unable to fetch schema with id %d" % (schema_id)
            raise SerializerError(err)

        return schema

    # Decoder support
    def _get_decoder_func(self, schema_id, payload):
        if schema_id in self.id_to_decoder_func:
            return self.id_to_decoder_func[schema_id]

        schema = self.get_schema(schema_id)

        curr_pos = payload.tell()

        if self.fast_avro:
            # try to use fast avro
            try:
                schema_dict = schema.to_json()
                payload.seek(curr_pos)
                decoder_func = lambda p: read_data(p, schema_dict)
                self.id_to_decoder_func[schema_id] = decoder_func
                return self.id_to_decoder_func[schema_id]
            except:
                payload.seek(curr_pos)
                pass

        avro_reader = avro.io.DatumReader(schema)

        def decoder(p):
            bin_decoder = avro.io.BinaryDecoder(p)
            return avro_reader.read(bin_decoder)

        self.id_to_decoder_func[schema_id] = decoder
        return self.id_to_decoder_func[schema_id]

    def decode_message(self, message):
        """
        Decode a message from kafka that has been encoded for use with
        the schema registry.
        """
        if len(message) <= 5:
            raise SerializerError("message is too small to decode")

        with ContextBytesIO(message) as payload:
            magic, schema_id = struct.unpack('>bI', payload.read(5))
            if magic != MAGIC_BYTE:
                raise SerializerError("message does not start with magic byte")
            decoder_func = self._get_decoder_func(schema_id, payload)
            return decoder_func(payload)



import re
import json
import http.server
from threading import Thread

from avroschemaserializer.schemaregistry.tests import data_gen

from avroschemaserializer.schemaregistry.serializers import Util


class ReqHandler(http.server.BaseHTTPRequestHandler):
    protocol_version = "HTTP/1.0"

    def do_GET(self):
        self.server._run_routes(self)

    def do_POST(self):
        self.server._run_routes(self)

    def log_message(self, format, *args):
        pass


class MockServer(http.server.HTTPServer, object):

    def __init__(self, *args, **kwargs):
        super(MockServer, self).__init__(*args, **kwargs)
        self.counts = { }
        self.schema_cache = { }
        self.all_routes = {
            'GET' : [
                (r"/schemas/ids/(\d+)", 'get_schema_by_id'),
                (r"/subjects/(\w+-\w+)/versions/latest", 'get_latest')
            ],
            'POST' : [
                (r"/subjects/(\w+-\w+)/versions", 'register'),
                (r"/subjects/(\w+-\w+)", 'get_version')
            ]
        }
        self.schema = data_gen.load_schema_file('basic_schema.avsc')

    def _send_response(self, resp, status, body):
        resp.send_response(status)
        resp.send_header("Content-Type", "application/json")
        resp.end_headers()
        resp.wfile.write(json.dumps(body).encode('utf-8'))

    def _create_error(self, msg, status=400, err_code=1):
        return (status, {
            "error_code" : err_code,
            "message" : msg
        })

    def _run_routes(self, req):
        self.add_count((req.command, req.path))
        routes = self.all_routes.get(req.command, [])
        for r in routes:
            m = re.match(r[0], req.path)
            if m:
                func = getattr(self, r[1])
                status, body = func(req, m.groups())
                ret = self._send_response(req, status, body)
                return ret

        # here means we got a bad req
        status, body = self._create_error("bad path specified")
        self._send_response(req, status, body)

    def get_schema_by_id(self, req, groups):
        result = {
            "schema" : self.schema
        }
        return (200, result)

    def _get_identity_schema(self, avro_schema):
        return json.dumps(avro_schema.to_json())

    def _get_schema_from_body(self, req):
        length = req.headers['content-length']
        data = req.rfile.read(int(length))
        data = json.loads(data.decode('utf-8'))
        schema = data.get("schema", None)

        if not schema:
            return None
        try:
            avro_schema = Util.parse_schema_from_string(schema)
            return self._get_identity_schema(avro_schema)
        except:
            return None

    def register(self, req, groups):
        schema_id = 1
        return (200, {'id' : schema_id })

    def get_version(self, req, groups):
        avro_schema = self._get_schema_from_body(req)

        if not avro_schema:
            return self._create_error("Invalid avro schema", 422, 42201)
        subject = groups[0]
        version = 1
        if version == -1:
            return self._create_error("Not found", 404)
        schema_id = 1

        result = {
            "schema" : avro_schema,
            "subject" : subject,
            "id" : schema_id,
            "version" : version
        }
        return (200, result)

    def get_latest(self, req, groups):
        subject = groups[0]
        result = {
            "schema" : self.schema,
            "subject" : subject,
            "id" : 1,
            "version" : 1
        }
        return (200, result)

    def add_count(self, path):
        if path not in self.counts:
            self.counts[path] = 0
        self.counts[path] += 1


class ServerThread(Thread):

    def __init__(self, port):
        Thread.__init__(self)
        self.server = None
        self.port = port

    def run(self):
        self.server = MockServer(('127.0.0.1', self.port), ReqHandler)
        self.server.serve_forever()

    def shutdown(self):
        if self.server:
            self.server.shutdown()
            self.server.socket.close()


if __name__ == '__main__':
    s = ServerThread(9001)
    s.start()



from setuptools import setup


def version():
    with open('VERSION') as f:
        return f.read().strip()


def reqs():
    return [
        line.strip() for line in open('requirements.txt') if not line.startswith('#')
    ]


setup(
    name             = 'avroschemaserializer',
    description      = 'Python 3 Confluent Schema Registry Client',
    version          = version(),
    license          = 'Apache 2.0',
    author           = 'avroschemaserializer',
    author_email     = 'gregory.h.chu@gmail.com',
    keywords         = 'avroschemaserializer schema registry schemaregistry confluent avro',
    install_requires = reqs(),
    tests_require    = ['mock'],
    url              = 'https://github.com/gregchu/python-serializers',
    classifiers      = [
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries',
    ],
    packages         = [
        'avroschemaserializer',
        'avroschemaserializer.schemaregistry',
        'avroschemaserializer.schemaregistry.serializers',
        'avroschemaserializer.schemaregistry.client',
        'avroschemaserializer.schemaregistry.tests'
    ],
)
